"""String constants for the OrientDB submodule."""


# Databases
BIOGRID = "biogrid"
CHEBI = "chebi"
CLINICAL_TRIALS = "clinical_trials"
CLINVAR = "clinvar"
DISGENET = "disgenet"
DRUGBANK = "drugbank"
ENSEMBL = "ensembl"
FLYBASE = "flybase"
GWAS_CATALOG = "gwas_catalog"
HGNC = "hgnc"
EXPRESSION_ATLAS = "expression_atlas"
INTACT = "intact"
IUPHAR = "iuphar"
KEGG = "kegg"
MGI = "mgi"
MIRTARBASE = "mirtarbase"
NCBI = "ncbi"
OFFSIDES = "offsides"
ONSIDES = "onsides"
NSIDES = "nsides"
PATHWAY_COMMONS = "pathway_commons"
PROTEIN_ATLAS = "protein_atlas"
REACTOME = "reactome"
RGD = "rgd"
SIDER = "sider"
STITCH = "stitch"
STRINGDB = "stringdb"
UNIPROT = "uniprot"

# Data Types
NODES = "nodes"
EDGES = "edges"
